--              Syntax Color
--
-- Modified slightly by R. Evans, 6-21-1997
--
-- These routines are used to display Euphoria program lines
-- in multiple colors. The editor (ed.ex) and the pretty printer (eprint.ex)
-- both include this file.

-- this file assumes that the following symbols have already
-- been defined:
--      SCREEN        - file/device number to write output to
--       NORMAL_COLOR - colors of various syntax classes
--      COMMENT_COLOR
--      KEYWORD_COLOR
--      BUILTIN_COLOR
--       STRING_COLOR
--      BRACKET_COLOR
--      INCABLE_COLOR - color of procedures and functions that are defined
--         TYPE_COLOR - color of defined types, probably the same as
--                    - BUILTIN_COLOR

include keywords.e

-- character classes
constant DIGIT = 1,
OTHER = 2,
LETTER  = 3,
BRACKET = 4,
QUOTE   = 5,
DASH = 6,
WHITE_SPACE = 7,
NEW_LINE = 8

sequence char_class

global sequence includes, obtypes -- needed to have dynamic naming conventions
includes = {} -- modifiable list of included funcs and procs
obtypes = {}    -- modifiable list of defined types

global procedure init_class()
    -- set up character classes for easier line scanning
    -- (assume no 0 char)
    char_class = repeat(OTHER, 255)
    
    char_class['a'..'z'] = LETTER
    char_class['A'..'Z'] = LETTER
    char_class['_'] = LETTER
    char_class['0'..'9'] = DIGIT
    char_class['['] = BRACKET
    char_class[']'] = BRACKET
    char_class['('] = BRACKET
    char_class[')'] = BRACKET
    char_class['{'] = BRACKET
    char_class['}'] = BRACKET
    char_class['\''] = QUOTE
    char_class['"'] = QUOTE
    char_class[' '] = WHITE_SPACE
    char_class['\t'] = WHITE_SPACE
    char_class['\n'] = NEW_LINE
    char_class['-'] = DASH
end procedure

sequence line  -- the line being processed
integer seg_start, seg_end -- start and end of current segment of line
integer color  -- the current color

procedure flush(integer new_color)
    -- if the color is changing, write out the current segment
    if new_color != color then
	if color != -1 then
	    text_color(color)
	    puts(SCREEN, line[seg_start..seg_end])
	    seg_start = seg_end + 1
	end if
	color = new_color
    end if
end procedure

function search_included(sequence input_line) -- Finalized
    sequence types, incls, found_includes, sub_name, file_name, validfnc, buffer
    integer quotes_found, file_num, grab
    object reading -- could be atom or sequence
    
    validfnc = {LETTER,DIGIT,DASH}
    
    incls = {}
    
    types = {}
    
    -- Extract the actual filename from the input_line
    
    file_name = {}
    
    grab = 0
    
    for char = 1 to length(input_line) by 1 do
	
	if (find(char_class[input_line[char]],validfnc) or input_line[char] = ':'
	    or input_line[char] = '\\' or input_line[char] = '.') then
	    if grab = 0 then
		grab = 1
	    end if
	else
	    grab = -1
	end if
	
	if grab = 1 then
	    file_name = file_name & input_line[char]
	end if
	
    end for
    
    file_num = open(file_name, "r") 
    
    -- Note: This used to use dir(), but that become unneeded
    
    if file_num = -1 then -- Is in the INCLUDES dir?
	file_num = open(getenv("EUDIR") & "\\INCLUDE\\" & file_name, "r")
    end if
    
    if file_num = -1 then -- if the file wasn't there, create a 'dummy' file
	buffer = {" "}
	
    else -- read it to buffer
	reading = {}
	buffer = {}
	
	while sequence(reading) do
	    reading = gets(file_num)
	    
	    if sequence(reading) then
		buffer = append(buffer, reading)
	    end if
	    
	end while
	
	close(file_num) -- done with the file
    end if
    
    -- Tear through the file line by line in search of procs and funcs
    -- This is rather redundant.  Oh well...
    
    for line = 1 to length(buffer) by 1 do
	if match("global procedure ", buffer[line]) and
	    (match("--", buffer[line]) > match("global procedure ",
	    buffer[line]) or match("--", buffer[line]) = 0) then
	    
	    -- only globals count from the included files
	    
	    quotes_found = 0
	    
	    -- Let's see if there are quotes or escape chars blocking it
	    for find_quote = 1 to match("global procedure ",
		buffer[line]) by 1 do
		if buffer[line][find_quote] = '"' then
		    if find_quote != 1 then
			if buffer[line][find_quote-1] != '\\' then
			    quotes_found = quotes_found + 1
			end if
		    else 
			quotes_found = quotes_found + 1
		    end if
		end if
	    end for
	    
	    if integer(quotes_found / 2) then -- It's good...
		
		sub_name = {}
		
		grab = 0
		
		-- Try to get just the proc or func name, not ('s or such
		for char2 = match("global procedure ",buffer[line]) + 17
		    to length(buffer[line]) by 1 do
		    
		    if (char_class[buffer[line][char2]] = LETTER or
			char_class[buffer[line][char2]] = DIGIT) and 
			grab = 0 then -- The beginning
			grab = 1
			
		    elsif (char_class[buffer[line][char2]] != LETTER and
			char_class[buffer[line][char2]] != DIGIT) and
			grab = 1 then -- The end
			grab = -1
			
		    end if
		    
		    if grab = 1 then -- If we're getting it, add to the name
			sub_name = sub_name & buffer[line][char2]
		    end if
		    
		end for
		
		incls = append(incls, sub_name) -- Add to the big list
		
	    end if
	    
	elsif match("global function ", buffer[line]) and
	    (match("--", buffer[line]) > match("global function ", buffer[line])
	    or match("--", buffer[line]) = 0) then
	    
	    -- only globals count from the included files
	    
	    quotes_found = 0
	    
	    -- Let's see if there are quotes or escape chars blocking it
	    for find_quote = 1 to match("global function ",
		buffer[line]) by 1 do
		if buffer[line][find_quote] = '"' then
		    if find_quote != 1 then
			if buffer[line][find_quote-1] != '\\' then
			    quotes_found = quotes_found + 1
			end if
		    else 
			quotes_found = quotes_found + 1
		    end if
		end if
	    end for
	    
	    if integer(quotes_found / 2) then -- It's good...
		
		sub_name = {}
		
		grab = 0
		
		-- Try to get just the proc or func name, not ('s or such
		for char2 = match("global function ",buffer[line]) + 16
		    to length(buffer[line]) by 1 do
		    
		    if (char_class[buffer[line][char2]] = LETTER or
			char_class[buffer[line][char2]] = DIGIT) and 
			grab = 0 then -- The beginning
			grab = 1
			
		    elsif (char_class[buffer[line][char2]] != LETTER and
			char_class[buffer[line][char2]] != DIGIT) and
			grab = 1 then -- The end
			grab = -1
			
		    end if
		    
		    if grab = 1 then -- If we're getting it, add to the name
			sub_name = sub_name & buffer[line][char2]
		    end if
		    
		end for
		
		incls = append(incls, sub_name) -- Add to the big list
		
	    end if
	    
	elsif match("global type ", buffer[line]) and
	    (match("--", buffer[line]) > match("global type ", buffer[line])
	    or match("--", buffer[line]) = 0) then
	    
	    -- only globals count from the included files
	    
	    quotes_found = 0
	    
	    -- Let's see if there are quotes or escape chars blocking it
	    for find_quote = 1 to match("global type ",
		buffer[line]) by 1 do
		if buffer[line][find_quote] = '"' then
		    if find_quote != 1 then
			if buffer[line][find_quote-1] != '\\' then
			    quotes_found = quotes_found + 1
			end if
		    else 
			quotes_found = quotes_found + 1
		    end if
		end if
	    end for
	    
	    if integer(quotes_found / 2) then -- It's good...
		
		sub_name = {}
		
		grab = 0
		
		-- Try to get just the proc or func name, not ('s or such
		for char2 = match("global type ",buffer[line]) + 12
		    to length(buffer[line]) by 1 do
		    
		    if (char_class[buffer[line][char2]] = LETTER or
			char_class[buffer[line][char2]] = DIGIT) and 
			grab = 0 then -- The beginning
			grab = 1
			
		    elsif (char_class[buffer[line][char2]] != LETTER and
			char_class[buffer[line][char2]] != DIGIT) and
			grab = 1 then -- The end
			grab = -1
			
		    end if
		    
		    if grab = 1 then -- If we're getting it, add to the name
			sub_name = sub_name & buffer[line][char2]
		    end if
		    
		end for
		
		types = append(types, sub_name) -- Add to the big list
		
	    end if
	    
	elsif match("include ", buffer[line]) and
	    (match("--", buffer[line]) > match("include ", buffer[line])
	    or match("--", buffer[line]) = 0) then
	    
	    quotes_found = 0
	    
	    for find_quote = 1 to match("include ", buffer[line]) by 1 do
		if buffer[line][find_quote] = '"' then
		    if find_quote != 1 then
			if buffer[line][find_quote-1] != '\\' then
			    quotes_found = quotes_found + 1
			end if
		    else 
			quotes_found = quotes_found + 1
		    end if
		end if
	    end for
	    
	    if integer(quotes_found / 2) then
		found_includes = search_included(buffer[line][match("include ",
		buffer[line])+8..length(buffer[line])])
		for add = 1 to length(found_includes[1]) by 1 do
		    incls = append(incls, found_includes[1][add])
		end for
		
		for add = 1 to length(found_includes[2]) by 1 do
		    types = append(types, found_includes[2][add])
		end for
	    end if
	end if
    end for
    
    return {incls,types}
    
end function -- search_included()

global procedure update_include_funcs(sequence buffer)
    integer grab, quotes_found
    sequence sub_name
    sequence found
    
    includes = {} -- Start all over in the includes business...
    obtypes = {} -- And in the types business...
    
    for line = 1 to length(buffer) by 1 do
	if match("include ", buffer[line]) and
	    (match("--", buffer[line]) > match("include ", buffer[line])
	    or match("--", buffer[line]) = 0) then
	    
	    quotes_found = 0
	    
	    for q = 1 to match("include ", buffer[line]) by 1 do
		if buffer[line][q] = '"' then
		    if q != 1 then
			if buffer[line][q-1] != '\\' then
			    quotes_found = quotes_found + 1
			end if
			
		    elsif q = 1 then 
			quotes_found = quotes_found + 1
			
		    end if
		end if
	    end for
	    
	    if integer(quotes_found / 2) then
		found = search_included(buffer[line][match("include ",
		buffer[line])+8..length(buffer[line])])
		for add = 1 to length(found[1]) by 1 do
		    includes = append(includes, found[1][add])
		end for
		
		for add = 1 to length(found[2]) by 1 do
		    obtypes = append(obtypes, found[2][add])
		end for
	    end if
	    
	elsif match("procedure ", buffer[line]) and
	    (match("--", buffer[line]) > match("procedure ", buffer[line]) or
	    match("--", buffer[line]) = 0) then
	    
	    quotes_found = 0
	    
	    for find_quote = 1 to match("procedure ", buffer[line]) by 1 do
		if buffer[line][find_quote] = '"' then
		    if find_quote != 1 then
			if buffer[line][find_quote-1] != '\\' then
			    quotes_found = quotes_found + 1
			end if
		    else 
			quotes_found = quotes_found + 1
		    end if
		end if
	    end for
	    
	    if integer(quotes_found / 2) then
		
		sub_name = {}
		
		grab = 0
		
		for char = match("procedure ", buffer[line]) + 10 to 
		    length(buffer[line]) by 1 do
		    if (char_class[buffer[line][char]] = LETTER or 
			char_class[buffer[line][char]] = DIGIT) then
			if grab = 0 then
			    grab = 1
			end if
			
		    else
			grab = -1
			
		    end if
		    
		    if grab = 1 then
			sub_name = sub_name & buffer[line][char]
		    end if
		    
		end for
		
		includes = append(includes, sub_name)
		
	    end if
	elsif match("function ", buffer[line]) and
	    (match("--", buffer[line]) > match("function ", buffer[line])
	    or match("--", buffer[line]) = 0) then
	    
	    quotes_found = 0
	    
	    for find_quote = 1 to match("function ", buffer[line]) by 1 do
		if buffer[line][find_quote] = '"' then
		    if find_quote != 1 then
			if buffer[line][find_quote-1] != '\\' then
			    quotes_found = quotes_found + 1
			end if
		    else 
			quotes_found = quotes_found + 1
		    end if
		end if
	    end for
	    
	    if quotes_found / 2 = floor(quotes_found / 2) then
		
		sub_name = {}
		
		grab = 0
		
		for char = match("function ", buffer[line]) + 9 to 
		    length(buffer[line]) by 1 do
		    if (char_class[buffer[line][char]] = LETTER or 
			char_class[buffer[line][char]] = DIGIT) then
			if grab = 0 then
			    grab = 1
			end if
		    
		    else
			grab = -1
		    
		    end if
		    
		    if grab = 1 then
			sub_name = sub_name & buffer[line][char]
		    end if
		
		end for
		
		includes = append(includes, sub_name)
	    
	    end if
	elsif match("type ", buffer[line]) and
	    (match("--", buffer[line]) > match("type ", buffer[line])
	    or match("--", buffer[line]) = 0) then
	    
	    quotes_found = 0
	    
	    for find_quote = 1 to match("type ", buffer[line]) by 1 do
		if buffer[line][find_quote] = '"' then
		    if find_quote != 1 then
			if buffer[line][find_quote-1] != '\\' then
			    quotes_found = quotes_found + 1
			end if
		    else 
			quotes_found = quotes_found + 1
		    end if
		end if
	    end for
	    
	    if quotes_found / 2 = floor(quotes_found / 2) then
		
		sub_name = {}
		
		grab = 0
		
		for char = match("type ", buffer[line]) + 5 to 
		    length(buffer[line]) by 1 do
		    if (char_class[buffer[line][char]] = LETTER or 
			char_class[buffer[line][char]] = DIGIT) then
			if grab = 0 then
			    grab = 1
			end if
		    
		    else
			grab = -1
		    
		    end if
		    
		    if grab = 1 then
			sub_name = sub_name & buffer[line][char]
		    end if
		
		end for
		
		obtypes = append(obtypes, sub_name)
	    
	    end if
	end if
    end for
end procedure -- update_include_funcs()

global procedure DisplayColorLine(sequence pline)
    -- Display a '\n'-terminated line with colors identifying the various
    -- parts of the Euphoria language.
    -- Each screen write has a lot of overhead, so we try to minimize
    -- the number of them by collecting consecutive characters of the
    -- same color into a 'segment' seg_start..seg_end.
    integer class, last, i, c, bracket_level
    sequence word
    
    line = pline
    color = -1 -- initially undefined
    bracket_level = 0
    seg_start = 1
    seg_end = 0
    
    while 1 do
	c = line[seg_end+1]
	class = char_class[c]
	
	if class = WHITE_SPACE then
	    seg_end = seg_end + 1 -- continue with same color
	    
	elsif class = LETTER then
	    last = length(line)-1
	    for j = seg_end + 2 to last do
		c = line[j]
		class = char_class[c]
		if class != LETTER then
		    if class != DIGIT then
			last = j - 1
			exit
		    end if
		end if
	    end for
	    word = line[seg_end+1..last]
	    if find(word, keywords) then
		flush(KEYWORD_COLOR)
	    elsif find(word, builtins) then
		flush(BUILTIN_COLOR)
	    elsif find(word, includes) then
		flush(INCABLE_COLOR)
	    elsif find(word, obtypes) then
		flush(TYPE_COLOR)
	    else
		flush(NORMAL_COLOR)
	    end if
	    seg_end = last
	    
	elsif class <= OTHER then -- DIGIT too
	    flush(NORMAL_COLOR)
	    seg_end = seg_end + 1
	    
	elsif class = BRACKET then
	    if find(c, "([{") then
		bracket_level = bracket_level + 1
	    end if
	    if bracket_level >= 1 and
		bracket_level <= length(BRACKET_COLOR) then
		flush(BRACKET_COLOR[bracket_level])
	    else
		flush(NORMAL_COLOR)
	    end if
	    if find(c, ")]}") then
		bracket_level = bracket_level - 1
	    end if
	    seg_end = seg_end + 1
	    
	elsif class = NEW_LINE then
	    exit  -- end of line
	    
	elsif class = DASH then
	    if line[seg_end+2] = '-' then
		flush(COMMENT_COLOR)
		seg_end = length(line)-1
		exit
	    end if
	    flush(NORMAL_COLOR)
	    seg_end = seg_end + 1
	    
	else  -- QUOTE
	    i = seg_end + 2
	    while i < length(line) do
		if line[i] = c then
		    i = i + 1
		    exit
		elsif line[i] = '\\' then
		    if i < length(line)-1 then
			i = i + 1 -- ignore escaped char
		    end if
		end if
		i = i + 1
	    end while
	    flush(STRING_COLOR)
	    seg_end = i - 1
	end if
    end while
    
    if color != -1 then
	text_color(color)
    end if
    puts(SCREEN, line[seg_start..seg_end])
end procedure


